@@ -1,4 +1,4 @@
-#!/usr/bin/perl -w
+#!/usr/bin/perl
 eval 'exec perl -S $0 "$@"'
     if 0;
 
@@ -6,7 +6,7 @@ eval 'exec perl -S $0 "$@"'
 
 pflogsumm.pl - Produce Postfix MTA logfile summary
 
-Copyright (C) 1998-2002 by James S. Seymour, Release 1.0.7.
+Copyright (C) 1998-2003 by James S. Seymour, Release 1.1.0.
 
 =head1 SYNOPSIS
 
@@ -14,7 +14,8 @@ Copyright (C) 1998-2002 by James S. Seymour, Release 1.0.7.
 	[--verp_mung[=<n>]] [--verbose_msg_detail] [--iso_date_time]
 	[-m|--uucp_mung] [-i|--ignore_case] [--smtpd_stats] [--mailq]
 	[--problems_first] [--rej_add_from] [--no_bounce_detail]
-	[--no_deferral_detail] [--no_reject_detail] [--zero_fill]
+	[--no_deferral_detail] [--no_reject_detail] [--no_no_msg_size]
+	[--no_smtpd_warnings] [--zero_fill] [--syslog_name=string]
 	[file1 [filen]]
 
     pflogsumm.pl -[help|version]
@@ -110,6 +111,25 @@ Copyright (C) 1998-2002 by James S. Seymour, Release 1.0.7.
                    See also: "-u" and "-h" for further report-limiting
                              options.
 
+    --no_no_msg_size
+
+		    Do not emit report on "Messages with no size data".
+
+		    Message size is reported only by the queue manager.
+		    The message may be delivered long-enough after the
+		    (last) qmgr log entry that the information is not in
+		    the log(s) processed by a particular run of
+		    pflogsumm.pl.  This throws off "Recipients by message
+		    size" and the total for "bytes delivered." These are
+		    normally reported by pflogsumm as "Messages with no
+		    size data."
+
+    --no_smtpd_warnings
+
+		    On a busy mail server, say at an ISP, SMTPD warnings
+		    can result in a rather sizeable report.  This option
+		    turns reporting them off.
+
     --problems_first
 
                    Emit "problems" reports (bounces, defers, warnings,
@@ -134,6 +154,19 @@ Copyright (C) 1998-2002 by James S. Seymour, Release 1.0.7.
                    reports.  For multiple-day reports: "per-hour" numbers
                    are daily averages (reflected in the report heading).
 
+    --syslog_name=name
+
+		   Set syslog_name to look for for Postfix log entries.
+
+		   By default, pflogsumm looks for entries in logfiles
+		   with a syslog name of "postfix," the default.
+		   If you've set a non-default "syslog_name" parameter
+		   in your Postfix configuration, use this option to
+		   tell pflogsumm what that is.
+
+		   See the discussion about the use of this option under
+		   "NOTES," below.
+
     -u <cnt>       top <cnt> to display in user reports. 0 == none.
 
                    See also: "-h" and "--no_*_detail" for further
@@ -254,6 +287,39 @@ Copyright (C) 1998-2002 by James S. Seymour, Release 1.0.7.
     may not be sorted correctly by-domain-by-user.  (Also depends on
     upstream MTA, I suspect.)
 
+    The "percent rejected" and "percent discarded" figures are only
+    approximations.  They are calculated as follows (example is for
+    "percent rejected"):
+
+	percent rejected =
+	
+	    (rejected / (delivered + rejected + discarded)) * 100
+
+    There are some issues with the use of --syslog_name.  The problem is
+    that, even with $syslog_name set, Postfix will sometimes still log
+    things with "postfix" as the syslog_name.  This is noted in
+    /etc/postfix/sample-misc.cf:
+
+	# Beware: a non-default syslog_name setting takes effect only
+	# after process initialization. Some initialization errors will be
+	# logged with the default name, especially errors while parsing
+	# the command line and errors while accessing the Postfix main.cf
+	# configuration file.
+
+    As a consequence, pflogsumm must always look for "postfix," in logs,
+    as well as whatever is supplied for syslog_name.
+
+    Where this becomes an issue is where people are running two or more
+    instances of Postfix, logging to the same file.  In such a case:
+
+	. Neither instance may use the default "postfix" syslog name
+	  and...
+
+	. Log entries that fall victim to what's described in
+	  sample-misc.cf will be reported under "postfix", so that if
+	  you're running pflogsumm twice, once for each syslog_name, such
+	  log entries will show up in each report.
+
     The Pflogsumm Home Page is at:
 
 	http://jimsun.LinxNet.com/postfix_contrib.html
@@ -297,7 +363,7 @@ use Date::Calc qw(Delta_DHMS);
 # ---End: SMTPD_STATS_SUPPORT---
 
 my $mailqCmd = "mailq";
-my $release = "1.0.7";
+my $release = "1.1.0";
 
 # Variables and constants used throughout pflogsumm
 use vars qw(
@@ -364,25 +430,22 @@ my (
 );
 $dayCnt = $smtpdConnCnt = $smtpdTotTime = 0;
 
-# Messages received and delivered per hour
-my @rcvPerHr = qw(0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0);
+# Init total messages delivered, rejected, and discarded
+$msgsDlvrd = $msgsRjctd = $msgsDscrdd = 0;
+
+# Init messages received and delivered per hour
+my @rcvPerHr = (0) x 24;
 my @dlvPerHr = @rcvPerHr;
 my @dfrPerHr = @rcvPerHr;	# defers per hour
 my @bncPerHr = @rcvPerHr;	# bounces per hour
 my @rejPerHr = @rcvPerHr;	# rejects per hour
 my $lastMsgDay = 0;
 
-# "doubly-sub-scripted array: cnt, total and max time per-hour
-# Gag - some things, Perl doesn't do well :-(
+# Init "doubly-sub-scripted array": cnt, total and max time per-hour
 my @smtpdPerHr;
-$smtpdPerHr[0]  = [0,0,0]; $smtpdPerHr[1]  = [0,0,0]; $smtpdPerHr[2]  = [0,0,0];
-$smtpdPerHr[3]  = [0,0,0]; $smtpdPerHr[4]  = [0,0,0]; $smtpdPerHr[5]  = [0,0,0];
-$smtpdPerHr[6]  = [0,0,0]; $smtpdPerHr[7]  = [0,0,0]; $smtpdPerHr[8]  = [0,0,0];
-$smtpdPerHr[9]  = [0,0,0]; $smtpdPerHr[10] = [0,0,0]; $smtpdPerHr[11] = [0,0,0];
-$smtpdPerHr[12] = [0,0,0]; $smtpdPerHr[13] = [0,0,0]; $smtpdPerHr[14] = [0,0,0];
-$smtpdPerHr[15] = [0,0,0]; $smtpdPerHr[16] = [0,0,0]; $smtpdPerHr[17] = [0,0,0];
-$smtpdPerHr[18] = [0,0,0]; $smtpdPerHr[19] = [0,0,0]; $smtpdPerHr[20] = [0,0,0];
-$smtpdPerHr[21] = [0,0,0]; $smtpdPerHr[22] = [0,0,0]; $smtpdPerHr[23] = [0,0,0];
+for (0 .. 23) {
+    $smtpdPerHr[$_]  = [0,0,0];
+}
 
 $progName = "pflogsumm.pl";
 $usageMsg =
@@ -390,7 +453,8 @@ $usageMsg =
        [--verp_mung[=<n>]] [--verbose_msg_detail] [--iso_date_time]
        [-m|--uucp_mung] [-i|--ignore_case] [--smtpd_stats] [--mailq]
        [--problems_first] [--rej_add_from] [--no_bounce_detail]
-       [--no_deferral_detail] [--no_reject_detail] [--zero_fill]
+       [--no_deferral_detail] [--no_reject_detail] [--no_no_msg_size]
+       [--no_smtpd_warnings] [--zero_fill] [--syslog_name=name]
        [file1 [filen]]
 
        $progName --[version|help]";
@@ -411,10 +475,13 @@ GetOptions(
     "no_bounce_detail"   => \$opts{'noBounceDetail'},
     "no_deferral_detail" => \$opts{'noDeferralDetail'},
     "no_reject_detail"   => \$opts{'noRejectDetail'},
+    "no_no_msg_size"     => \$opts{'noNoMsgSize'},
+    "no_smtpd_warnings"  => \$opts{'noSMTPDWarnings'},
     "problems_first"     => \$opts{'pf'},
     "q"                  => \$opts{'q'},
     "rej_add_from"       => \$opts{'rejAddFrom'},
     "smtpd_stats"        => \$opts{'smtpdStats'},
+    "syslog_name=s"      => \$opts{'syslogName'},
     "u=i"                => \$opts{'u'},
     "verbose_msg_detail" => \$opts{'verbMsgDetail'},
     "verp_mung:i"        => \$opts{'verpMung'},
@@ -425,6 +492,7 @@ GetOptions(
 # internally: 0 == none, undefined == -1 == all
 $opts{'h'} = -1 unless(defined($opts{'h'}));
 $opts{'u'} = -1 unless(defined($opts{'u'}));
+my $syslogName = $opts{'syslogName'}? $opts{'syslogName'} : "postfix";
 
 if(defined($opts{'help'})) {
     print "$usageMsg\n";
@@ -444,23 +512,21 @@ $dateStr = get_datestr($opts{'d'}) if(defined($opts{'d'}));
 
 while(<>) {
     next if(defined($dateStr) && ! /^$dateStr/o);
-    s/: \[ID \d+\s+[^\]]+\] /: /o;	# lose "[ID nnnnnn some.thing]" stuff
-    unless((($msgMonStr, $msgDay, $msgTimeStr, $cmd, $qid) =
-	m#^(...)\s+(\d+)\s(..:..:..)\s\S+\s(?:vmailer|postfix)[-/]([^\[:]*).*?: ([^:]+)#o) == 5)
+    s/: \[ID \d+ [^\]]+\] /: /o;	# lose "[ID nnnnnn some.thing]" stuff
+    my $logRmdr;
+    next unless((($msgMonStr, $msgDay, $msgHr, $msgMin, $msgSec, $logRmdr) =
+	/^(...) +(\d+) (..):(..):(..) \S+ (.+)$/o) == 6);
+    unless((($cmd, $qid) = $logRmdr =~ m#^(?:vmailer|postfix|$syslogName)/([^\[:]*).*?: ([^:\s]+)#o) == 2 ||
+           (($cmd, $qid) = $logRmdr =~ m#^((?:vmailer|postfix)(?:-script)?)(?:\[\d+\])?: ([^:\s]+)#o) == 2)
     {
-	unless((($msgMonStr, $msgDay, $msgTimeStr, $cmd, $qid) =
-	    m#^(...)\s+(\d+)\s(..:..:..)\s\S+\s(vmailer|postfix[^\[:]*).*?: ([^:]+)#o) == 5)
-	{
-#		print UNPROCD "$_";
-		next;
-	}
+	#print UNPROCD "$_";
+	next;
     }
     chomp;
 
     # snatch out log entry date & time
-    ($msgHr, $msgMin, $msgSec) = split(/:/, $msgTimeStr);
     $msgMon = $monthNums{$msgMonStr};
-    $msgYr = $thisYr; --$msgYr if($msgMon > $thisMon);
+    $msgYr = ($msgMon > $thisMon? $thisYr - 1 : $thisYr);
 
     # the following test depends on one getting more than one message a
     # month--or at least that successive messages don't arrive on the
@@ -475,76 +541,78 @@ while(<>) {
     }
 
     # regexp rejects happen in "cleanup"
-    if((my($rejTyp, $rejSubTyp, $rejReas, $rejRmdr) =
-	/^.*\/(cleanup)\[.*?\b(reject|warning|hold|discard): (header|body) (.*)$/o) == 4)
+    if($cmd eq "cleanup" && (my($rejSubTyp, $rejReas, $rejRmdr) = $logRmdr =~
+	/\/cleanup\[\d+\]: .*?\b(reject|warning|hold|discard): (header|body) (.*)$/o) == 3)
     {
 	$rejRmdr =~ s/( from \S+?)?; from=<.*$//o unless($opts{'verbMsgDetail'});
 	$rejRmdr = string_trimmer($rejRmdr, 64, $opts{'verbMsgDetail'});
 	if($rejSubTyp eq "reject") {
-	    ++$rejects{$rejTyp}{$rejReas}{$rejRmdr};
+	    ++$rejects{$cmd}{$rejReas}{$rejRmdr};
 	    ++$msgsRjctd;
 	} elsif($rejSubTyp eq "warning") {
-	    ++$warns{$rejTyp}{$rejReas}{$rejRmdr};
+	    ++$warns{$cmd}{$rejReas}{$rejRmdr};
 	    ++$msgsWrnd;
 	} elsif($rejSubTyp eq "hold") {
-	    ++$holds{$rejTyp}{$rejReas}{$rejRmdr};
+	    ++$holds{$cmd}{$rejReas}{$rejRmdr};
 	    ++$msgsHld;
 	} elsif($rejSubTyp eq "discard") {
-	    ++$discards{$rejTyp}{$rejReas}{$rejRmdr};
+	    ++$discards{$cmd}{$rejReas}{$rejRmdr};
 	    ++$msgsDscrdd;
 	}
 	++$rejPerHr[$msgHr];
 	++${$msgsPerDay{$revMsgDateStr}}[4];
     } elsif($qid eq 'warning') {
-	(my $warnReas = $_) =~ s/^.*warning: //o;
+	(my $warnReas = $logRmdr) =~ s/^.*warning: //o;
 	$warnReas = string_trimmer($warnReas, 66, $opts{'verbMsgDetail'});
-	++$warnings{$cmd}{$warnReas};
+	unless($cmd eq "smtpd" && $opts{'noSMTPDWarnings'}) {
+	    ++$warnings{$cmd}{$warnReas};
+	}
     } elsif($qid eq 'fatal') {
-	(my $fatalReas = $_) =~ s/^.*fatal: //o;
+	(my $fatalReas = $logRmdr) =~ s/^.*fatal: //o;
 	$fatalReas = string_trimmer($fatalReas, 66, $opts{'verbMsgDetail'});
 	++$fatals{$cmd}{$fatalReas};
     } elsif($qid eq 'panic') {
-	(my $panicReas = $_) =~ s/^.*panic: //o;
+	(my $panicReas = $logRmdr) =~ s/^.*panic: //o;
 	$panicReas = string_trimmer($panicReas, 66, $opts{'verbMsgDetail'});
 	++$panics{$cmd}{$panicReas};
     } elsif($qid eq 'reject') {
-	proc_smtpd_reject($_, \%rejects, \$msgsRjctd, \$rejPerHr[$msgHr],
+	proc_smtpd_reject($logRmdr, \%rejects, \$msgsRjctd, \$rejPerHr[$msgHr],
 			  \${$msgsPerDay{$revMsgDateStr}}[4]);
     } elsif($qid eq 'reject_warning') {
-	proc_smtpd_reject($_, \%warns, \$msgsWrnd, \$rejPerHr[$msgHr],
+	proc_smtpd_reject($logRmdr, \%warns, \$msgsWrnd, \$rejPerHr[$msgHr],
 			  \${$msgsPerDay{$revMsgDateStr}}[4]);
     } elsif($qid eq 'hold') {
-	proc_smtpd_reject($_, \%holds, \$msgsHld, \$rejPerHr[$msgHr],
+	proc_smtpd_reject($logRmdr, \%holds, \$msgsHld, \$rejPerHr[$msgHr],
 			  \${$msgsPerDay{$revMsgDateStr}}[4]);
     } elsif($qid eq 'discard') {
-	proc_smtpd_reject($_, \%discards, \$msgsDscrdd, \$rejPerHr[$msgHr],
+	proc_smtpd_reject($logRmdr, \%discards, \$msgsDscrdd, \$rejPerHr[$msgHr],
 			  \${$msgsPerDay{$revMsgDateStr}}[4]);
     } elsif($cmd eq 'master') {
-	++$masterMsgs{(split(/^.*master.*: /))[1]};
+	++$masterMsgs{(split(/^.*master.*: /, $logRmdr))[1]};
     } elsif($cmd eq 'smtpd') {
-	if(/\[\d+\]: \w+: client=/o) {
+	if($logRmdr =~ /\[\d+\]: \w+: client=(.+?)(,|$)/o) {
 	    #
 	    # Warning: this code in two places!
 	    #
 	    ++$rcvPerHr[$msgHr];
 	    ++${$msgsPerDay{$revMsgDateStr}}[0];
 	    ++$msgsRcvd;
-	    ++$rcvdMsg{$qid};	# quick-set a flag
-	} elsif(my($rejSubTyp) = /\[\d+\]: \w+: (reject(?:_warning)?|hold|discard): /o) {
+	    $rcvdMsg{$qid} = gimme_domain($1);	# Whence it came
+	} elsif(my($rejSubTyp) = $logRmdr =~ /\[\d+\]: \w+: (reject(?:_warning)?|hold|discard): /o) {
 	    if($rejSubTyp eq 'reject') {
-		proc_smtpd_reject($_, \%rejects, \$msgsRjctd,
+		proc_smtpd_reject($logRmdr, \%rejects, \$msgsRjctd,
 				  \$rejPerHr[$msgHr],
 				  \${$msgsPerDay{$revMsgDateStr}}[4]);
 	    } elsif($rejSubTyp eq 'reject_warning') {
-		proc_smtpd_reject($_, \%warns, \$msgsWrnd,
+		proc_smtpd_reject($logRmdr, \%warns, \$msgsWrnd,
 				  \$rejPerHr[$msgHr],
 				  \${$msgsPerDay{$revMsgDateStr}}[4]);
 	    } elsif($rejSubTyp eq 'hold') {
-		proc_smtpd_reject($_, \%holds, \$msgsHld,
+		proc_smtpd_reject($logRmdr, \%holds, \$msgsHld,
 				  \$rejPerHr[$msgHr],
 				  \${$msgsPerDay{$revMsgDateStr}}[4]);
 	    } elsif($rejSubTyp eq 'discard') {
-		proc_smtpd_reject($_, \%discards, \$msgsDscrdd,
+		proc_smtpd_reject($logRmdr, \%discards, \$msgsDscrdd,
 				  \$rejPerHr[$msgHr],
 				  \${$msgsPerDay{$revMsgDateStr}}[4]);
 	    }
@@ -552,12 +620,12 @@ while(<>) {
 # ---Begin: SMTPD_STATS_SUPPORT---
 	else {
 	    next unless(defined($opts{'smtpdStats'}));
-	    if(/: connect from /o) {
-		/\/smtpd\[(\d+)\]: /o;
+	    if($logRmdr =~ /: connect from /o) {
+		$logRmdr =~ /\/smtpd\[(\d+)\]: /o;
 		@{$connTime{$1}} =
 		    ($msgYr, $msgMon + 1, $msgDay, $msgHr, $msgMin, $msgSec);
-	    } elsif(/: disconnect from /o) {
-		my ($pid, $hostID) = /\/smtpd\[(\d+)\]: disconnect from (.+)$/o;
+	    } elsif($logRmdr =~ /: disconnect from /o) {
+		my ($pid, $hostID) = $logRmdr =~ /\/smtpd\[(\d+)\]: disconnect from (.+)$/o;
 		if(exists($connTime{$pid})) {
 		    $hostID = gimme_domain($hostID);
 		    my($d, $h, $m, $s) = Delta_DHMS(@{$connTime{$pid}},
@@ -593,7 +661,7 @@ while(<>) {
 # ---End: SMTPD_STATS_SUPPORT---
     } else {
 	my $toRmdr;
-	if((($addr, $size) = /from=<([^>]*)>, size=(\d+)/o) == 2)
+	if((($addr, $size) = $logRmdr =~ /from=<([^>]*)>, size=(\d+)/o) == 2)
 	{
 	    next if($msgSizes{$qid});	# avoid double-counting!
 	    if($addr) {
@@ -610,7 +678,12 @@ while(<>) {
 	    push(@{$msgDetail{$qid}}, $addr) if($opts{'e'});
 	    # Avoid counting forwards
 	    if($rcvdMsg{$qid}) {
-		(my $domAddr = $addr) =~ s/^[^@]+\@//o;	# get domain only
+		# Get the domain out of the sender's address.  If there is
+		# none: Use the client hostname/IP-address
+		my $domAddr;
+		unless((($domAddr = $addr) =~ s/^[^@]+\@(.+)$/$1/o) == 1) {
+		    $domAddr = $rcvdMsg{$qid} eq "pickup"? $addr : $rcvdMsg{$qid};
+		}
 		++$sendgDomCnt
 		    unless(${$sendgDom{$domAddr}}[$msgCntI]);
 		++${$sendgDom{$domAddr}}[$msgCntI];
@@ -622,8 +695,8 @@ while(<>) {
 		delete($rcvdMsg{$qid});		# limit hash size
 	    }
 	}
-	elsif((($addr, $relay, $delay, $status, $toRmdr) =
-		/to=<([^>]*)>, relay=([^,]+), delay=([^,]+), status=([^ ]+)(.*)$/o) >= 4)
+	elsif((($addr, $relay, $delay, $status, $toRmdr) = $logRmdr =~
+		/to=<([^>]*)>, (?:orig_to=<[^>]*>, )?relay=([^,]+), delay=([^,]+), status=(\S+)(.*)$/o) >= 4)
 	{
 
 	    if($opts{'m'} && $addr =~ /^(.*!)*([^!]+)!([^!@]+)@([^\.]+)$/o) {
@@ -659,14 +732,14 @@ while(<>) {
 		} else {
 		    ${$recipDom{$domAddr}}[$msgSizeI] += 0;
 		    ${$recipUser{$addr}}[$msgSizeI] += 0;
-		    $noMsgSize{$qid} = $addr;
+		    $noMsgSize{$qid} = $addr unless($opts{'noNoMsgSize'});
 		    push(@{$msgDetail{$qid}}, "(sender not in log)") if($opts{'e'});
 		    # put this back later? mebbe with -v?
 		    # msg_warn("no message size for qid: $qid");
 		}
 		push(@{$msgDetail{$qid}}, $addr) if($opts{'e'});
 	    } elsif($status eq 'deferred') {
-		my ($deferredReas) = /, status=deferred \(([^\)]+)/o;
+		my ($deferredReas) = $logRmdr =~ /, status=deferred \(([^\)]+)/o;
 		unless(defined($opts{'verbMsgDetail'})) {
 		    $deferredReas = said_string_trimmer($deferredReas, 65);
 		    $deferredReas =~ s/^\d{3} //o;
@@ -684,7 +757,7 @@ while(<>) {
 		    ${$recipDom{$domAddr}}[$msgDlyMaxI] = $delay
 		}
 	    } elsif($status eq 'bounced') {
-		my ($bounceReas) = /, status=bounced \((.+)\)/o;
+		my ($bounceReas) = $logRmdr =~ /, status=bounced \((.+)\)/o;
 		unless(defined($opts{'verbMsgDetail'})) {
 		    $bounceReas = said_string_trimmer($bounceReas, 66);
 		    $bounceReas =~ s/^\d{3} //o;
@@ -697,19 +770,20 @@ while(<>) {
 #		print UNPROCD "$_\n";
 	    }
 	}
-	elsif($cmd eq 'pickup' && /: (sender|uid)=/o) {
+	elsif($cmd eq 'pickup' && $logRmdr =~ /: (sender|uid)=/o) {
 	    #
 	    # Warning: this code in two places!
 	    #
 	    ++$rcvPerHr[$msgHr];
 	    ++${$msgsPerDay{$revMsgDateStr}}[0];
 	    ++$msgsRcvd;
-	    ++$rcvdMsg{$qid};	# quick-set a flag
+	    $rcvdMsg{$qid} = "pickup";	# Whence it came
 	}
 	elsif($cmd eq 'smtp') {
-	    if(/.* connect to ([^:]+): ([^;]+); address [\d\.]+ port.*$/o) {
+	    # Was an IPv6 problem here
+	    if($logRmdr =~ /.* connect to (\S+?): ([^;]+); address \S+ port.*$/o) {
 		++$smtpMsgs{lc($2)}{$1};
-	    } elsif(/.* connect to ([^[]+)\[[\d\.]+]: (.+) \(port \d+\)$/o) {
+	    } elsif($logRmdr =~ /.* connect to ([^[]+)\[\S+?\]: (.+?) \(port \d+\)$/o) {
 		++$smtpMsgs{lc($2)}{$1};
 	    } else {
 #		print UNPROCD "$_\n";
@@ -726,6 +800,14 @@ while(<>) {
 #close(UNPROCD) ||
 #    die "problem closing \"unprocessed\": $!\n";
 
+# Calculate percentage of messages rejected and discarded
+my $msgsRjctdPct = 0;
+my $msgsDscrddPct = 0;
+if(my $msgsTotal = $msgsDlvrd + $msgsRjctd + $msgsDscrdd) {
+    $msgsRjctdPct = int(($msgsRjctd/$msgsTotal) * 100);
+    $msgsDscrddPct = int(($msgsDscrdd/$msgsTotal) * 100);
+}
+
 if(defined($dateStr)) {
     print "Postfix log summaries for $dateStr\n";
 }
@@ -739,10 +821,10 @@ printf " %6d%s  deferred", adj_int_units($msgsDfrd);
 printf "  (%d%s deferrals)", adj_int_units($msgsDfrdCnt) if($msgsDfrdCnt);
 print "\n";
 printf " %6d%s  bounced\n", adj_int_units($msgsBncd);
-printf " %6d%s  rejected\n", adj_int_units($msgsRjctd);
+printf " %6d%s  rejected (%d%%)\n", adj_int_units($msgsRjctd), $msgsRjctdPct;
 printf " %6d%s  reject warnings\n", adj_int_units($msgsWrnd);
 printf " %6d%s  held\n", adj_int_units($msgsHld);
-printf " %6d%s  discarded\n", adj_int_units($msgsDscrdd);
+printf " %6d%s  discarded (%d%%)\n", adj_int_units($msgsDscrdd), $msgsDscrddPct;
 print "\n";
 printf " %6d%s  bytes received\n", adj_int_units($sizeRcvd);
 printf " %6d%s  bytes delivered\n", adj_int_units($sizeDlvrd);
@@ -956,9 +1038,10 @@ sub print_user_data {
 	$dottedLine = "\n" . "-" x length($title);
     }
     printf "\n$title$dottedLine\n";
-    foreach (reverse sort { ${$hashRef->{$a}}[$index] <=>
-	                    ${$hashRef->{$b}}[$index] }
-	keys(%$hashRef))
+    foreach (map { $_->[0] }
+	     sort { $b->[1] <=> $a->[1] || $a->[2] cmp $b->[2] }
+	     map { [ $_, $hashRef->{$_}[$index], normalize_host($_) ] }
+	     (keys(%$hashRef)))
     {
 	printf " %6d%s  %s\n", adj_int_units(${$hashRef->{$_}}[$index]), $_;
 	last if --$cnt == 0;
@@ -1139,15 +1222,25 @@ sub walk_nested_hash {
 
     if(ref($hashVal) eq 'HASH') {
 	foreach (sort keys %$hashRef) {
-	    print "$indents$_\n";
+	    print "$indents$_";
+	    # If the next hash is finally the data, total the
+	    # counts for the report and print
+	    my $hashVal2 = (each(%{$hashRef->{$_}}))[1];
+	    keys(%{$hashRef->{$_}});	# "reset" hash iterator
+	    unless(ref($hashVal2) eq 'HASH') {
+		my $cnt = 0;
+		$cnt += $_ foreach (values %{$hashRef->{$_}});
+		print " (total: $cnt)";
+	    }
+	    print "\n";
 	    walk_nested_hash($hashRef->{$_}, $level);
 	}
     } else {
 	really_print_hash_by_cnt_vals($hashRef, 0, $indents);
-#	print "\n"
     }
 }
 
+
 # print per-message info in excruciating detail :-)
 sub print_detailed_msg_data {
     use vars '$hashRef';
@@ -1172,15 +1265,33 @@ sub print_detailed_msg_data {
 }
 
 # *really* print hash contents sorted by numeric values in descending
-# order (i.e.: highest first) :-)
+# order (i.e.: highest first), then by IP/addr, in ascending order.
 sub really_print_hash_by_cnt_vals {
     my($hashRef, $cnt, $indents) = @_;
 
-    foreach (reverse sort { $hashRef->{$a} <=> $hashRef->{$b} }
-	keys(%$hashRef))
+    foreach (map { $_->[0] }
+	     sort { $b->[1] <=> $a->[1] || $a->[2] cmp $b->[2] }
+	     map { [ $_, $hashRef->{$_}, normalize_host($_) ] }
+	     (keys(%$hashRef)))
     {
-	printf "$indents%6d%s  %s\n", adj_int_units($hashRef->{$_}), $_;
-	last if --$cnt == 0;
+        printf "$indents%6d%s  %s\n", adj_int_units($hashRef->{$_}), $_;
+        last if --$cnt == 0;
+    }
+}
+
+# Normalize IP addr or hostname
+# (Note: Makes no effort to normalize IPv6 addrs.  Just returns them
+# as they're passed-in.)
+sub normalize_host {
+    # For IP addrs and hostnames: lop off possible " (user@dom.ain)" bit
+    my $norm1 = (split(/\s/, $_[0]))[0];
+
+    if((my @octets = ($norm1 =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/o)) == 4) {
+	# Dotted-quad IP address
+	return(pack('C4', @octets));
+    } else {
+	# Possibly hostname or user@dom.ain
+	return(join( '', map { lc $_ } reverse split /[.@]/, $norm1 ));
     }
 }
 
@@ -1270,23 +1381,27 @@ sub get_datestr {
     return sprintf("%s %2d", $monthNames[$t_mon], $t_mday);
 }
 
-# if there's a real domain: uses that.  Otherwise uses the first
-# three octets of the IP addr.  (In the latter case: usually pretty
-# safe to assume it's a dialup with a class C IP addr.)  Lower-
-# cases returned domain name.
+# if there's a real domain: uses that.  Otherwise uses the IP addr.
+# Lower-cases returned domain name.
+#
+# Optional bit of code elides the last octet of an IPv4 address.
+# (In case one wants to assume an IPv4 addr. is a dialup or other
+# dynamic IP address in a /24.)
+# Does nothing interesting with IPv6 addresses.
 sub gimme_domain {
     $_ = $_[0];
     my($domain, $ipAddr);
  
     # split domain/ipaddr into separates
-    unless((($domain, $ipAddr) = /^([^\[]+)\[([^\]]+)\]:?\s*$/o) == 2) {
+    # newer versions of Postfix have them "dom.ain[i.p.add.ress]"
+    # older versions of Postfix have them "dom.ain/i.p.add.ress"
+    unless((($domain, $ipAddr) = /^([^\[]+)\[([^\]]+)\]/o) == 2 ||
+           (($domain, $ipAddr) = /^([^\/]+)\/([0-9a-f.:]+)/oi) == 2) {
 	# more exhaustive method
-        ($domain, $ipAddr) = /^([^\[\(]+)[\[\(]([^\]\)]+)[\]\)]:?\s*$/o;
+        ($domain, $ipAddr) = /^([^\[\(\/]+)[\[\(\/]([^\]\)]+)[\]\)]?:?\s*$/o;
     }
  
-#    print STDERR "dbg: in=\"$_\", domain=\"$domain\", ipAddr=\"$ipAddr\"\n";
-    # now re-order "mach.host.dom"/"mach.host.do.co" to
-    # "host.dom.mach"/"host.do.co.mach"
+    # "mach.host.dom"/"mach.host.do.co" to "host.dom"/"host.do.co"
     if($domain eq 'unknown') {
         $domain = $ipAddr;
 	# For identifying the host part on a Class C network (commonly
@@ -1377,15 +1492,21 @@ sub proc_smtpd_reject {
     # This could get real ugly!
 
     # First: get everything following the "reject: ", etc. token
+    # Was an IPv6 problem here
     ($rejTyp, $rejFrom, $rejRmdr) = 
-	($logLine =~ /^.* \b(?:reject(?:_warning)?|hold|discard): ([^ ]+) from ([^:]+): (.*)$/o);
+	($logLine =~ /^.* \b(?:reject(?:_warning)?|hold|discard): (\S+) from (\S+?): (.*)$/o);
 
     # Next: get the reject "reason"
     $rejReas = $rejRmdr;
     unless(defined($opts{'verbMsgDetail'})) {
-	if($rejTyp eq "RCPT") {	# special treatment :-(
+	if($rejTyp eq "RCPT" || $rejTyp eq "DATA" || $rejTyp eq "CONNECT") {	# special treatment :-(
+	    # If there are "<>"s immediately following the reject code, that's
+	    # an email address or HELO string.  There can be *anything* in
+	    # those--incl. stuff that'll screw up subsequent parsing.  So just
+	    # get rid of it right off.
+	    $rejReas =~ s/^(\d{3} <).*?(>:)/$1$2/o;
 	    $rejReas =~ s/^(?:.*?[:;] )(?:\[[^\]]+\] )?([^;,]+)[;,].*$/$1/o;
-	    $rejReas =~ s/^(Sender address rejected: [^:]+):.*$/$1/o;
+	    $rejReas =~ s/^((?:Sender|Recipient) address rejected: [^:]+):.*$/$1/o;
 	    $rejReas =~ s/(Client host|Sender address) .+? blocked/blocked/o;
 	} elsif($rejTyp eq "MAIL") {	# *more* special treatment :-( grrrr...
 	    $rejReas =~ s/^\d{3} (?:<.+>: )?([^;:]+)[;:]?.*$/$1/o;
@@ -1394,10 +1515,17 @@ sub proc_smtpd_reject {
 	}
     }
 
-    # Snag out the "from" and "to" addrs...
-    unless((($from, $to) = $rejRmdr =~ m/from=<(.*?)>\s+to=<(.*?)>/o) == 2) {
-	($to, $from) = $rejRmdr =~ m/\d{3} <(.*?)>: User unknown .+from=<(.*?)> proto=/o;
-    }
+    # Snag recipient address
+    # Second expression is for unknown recipient--where there is no
+    # "to=<mumble>" field, third for pathological case where recipient
+    # field is unterminated, forth when all else fails.
+    (($to) = $rejRmdr =~ /to=<([^>]+)>/o) ||
+	(($to) = $rejRmdr =~ /\d{3} <([^>]+)>: User unknown /o) ||
+	(($to) = $rejRmdr =~ /to=<(.*?)(?:[, ]|$)/o) ||
+	($to = "<>");
+
+    # Snag sender address
+    (($from) = $rejRmdr =~ /from=<([^>]+)>/o) || ($from = "<>");
 
     if(defined($from)) {
 	$rejAddFrom = $opts{'rejAddFrom'};
@@ -1409,19 +1537,30 @@ sub proc_smtpd_reject {
 	# Sender address rejected: Domain not found
 	# Sender address rejected: need fully-qualified address
 	++$rejects->{$rejTyp}{$rejReas}{$from};
-    } elsif($rejReas =~ m/^Recipient address rejected:/o) {
+    } elsif($rejReas =~ m/^(Recipient address rejected:|User unknown( |$))/o) {
 	# Recipient address rejected: Domain not found
 	# Recipient address rejected: need fully-qualified address
-	++$rejects->{$rejTyp}{$rejReas}{$to};
+	# User unknown (in local/relay recipient table)
+	#++$rejects->{$rejTyp}{$rejReas}{$to};
+	my $rejData = $to;
+	if($rejAddFrom) {
+	    $rejData .= "  (" . ($from? $from : gimme_domain($rejFrom)) . ")";
+	}
+	++$rejects->{$rejTyp}{$rejReas}{$rejData};
     } elsif($rejReas =~ s/^.*?\d{3} (Improper use of SMTP command pipelining);.*$/$1/o) {
-	my ($src) = $logLine =~ /^.+? from ([^:]+):.*$/o;
+	# Was an IPv6 problem here
+	my ($src) = $logLine =~ /^.+? from (\S+?):.*$/o;
 	++$rejects->{$rejTyp}{$rejReas}{$src};
     } elsif($rejReas =~ s/^.*?\d{3} (Message size exceeds fixed limit);.*$/$1/o) {
 	my $rejData = gimme_domain($rejFrom);
 	$rejData .= "  ($from)" if($rejAddFrom);
 	++$rejects->{$rejTyp}{$rejReas}{$rejData};
+    } elsif($rejReas =~ s/^.*?\d{3} (Server configuration (?:error|problem));.*$/(Local) $1/o) {
+	my $rejData = gimme_domain($rejFrom);
+	$rejData .= "  ($from)" if($rejAddFrom);
+	++$rejects->{$rejTyp}{$rejReas}{$rejData};
     } else {
-#	    print STDERR "dbg: unknown reject reason $rejReas !\n\n";
+#	print STDERR "dbg: unknown reject reason $rejReas !\n\n";
 	my $rejData = gimme_domain($rejFrom);
 	$rejData .= "  ($from)" if($rejAddFrom);
 	++$rejects->{$rejTyp}{$rejReas}{$rejData};
